/*
 *	Copyright (C) 2007 Telechips Inc.
 *	TCC7901 scan keyboard driver
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/miscdevice.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/arch/tcc79x.h>

#define DEVICE_NAME	"btn"

#define MAX_KC_TIMES  8
#define IRQ_ADC	30

static int BtnMajor = 0;
static int k_scan;
typedef struct {
	int s_scancode;
	int e_scancode;
	char vkcode;
}SCANCODE_MAPPING;
SCANCODE_MAPPING key_mapping[]=
{
	{0x29, 0x2c, 1},		// A2    
	{0x34, 0x38, 2},		// A3
	{0x02, 0x06, 3},		// A4
	{0x0a, 0x0d, 4},		// A5
	{0x22, 0x24, 5},		// A6
	{0x74, 0x78, 6},		// B1
	{0x1c, 0x1f, 7},		// B2
	{0x66, 0x6a, 8},		// B3
	{0x56, 0x5a, 9},		// B4
	{0x11, 0x15, 10},		// B5
	{0x46, 0x4a, 11},		// B6
};

static void delay(int cnt)
{
	while(cnt --);
}

/**************************************************
 *	Get system btn A/D value
 *	Return: < 0x7f (10bits A/D)
 *************************************************/
// use A/D converter to read the scancode
static int GetScanCode(void)
{
	int c;
	
	do{
		HwADCCON &= ~(1<<4);
		c = HwADDATA;
	}while (!(c & 1));
	//printk("GetScanCode:%x\n", c);
	
	delay(1000);
	
	HwADCCON |= (1<<4);
	return c>>4;
}

// return the keycode, return -1 if no key matched
static int GetKeyCodeByScanCode(int kc)
{
	int i;
	
	for (i = 0;i < sizeof(key_mapping)/sizeof(SCANCODE_MAPPING);i++)
		if ((kc >= key_mapping[i].s_scancode) && (kc <= key_mapping[i].e_scancode)){
			return key_mapping[i].vkcode;
		}
	return 0;
}

static irqreturn_t btn_handler(int irq, void *dev_id)
{
	int kc[MAX_KC_TIMES];
	int nKey[MAX_KC_TIMES];
	int i;
	char nFlag;
	
	//printk("Keypad interrupt\n");
	
	HwADCCFG &= ~(1<<3); // Disable ADC IRQ
	HwADCCON |= (1<<4);
	
	for(i=0; i<MAX_KC_TIMES; i++){
		kc[i] = GetScanCode();
		nKey[i] = GetKeyCodeByScanCode(kc[i]);
	}
	//if(kc[0] < 0x7a)
		//printk("KEY:%x, RAW KC:%x,%x,%x,%x,%x,%x\n", nKey[0],kc[0],kc[1],kc[2],kc[3],kc[4],kc[5]);
	
	nFlag = 1;
	for (i=1; i<MAX_KC_TIMES; i++)
		if (nKey[i] != nKey[i-1]){
			nFlag = 0;
			break;
		}
	
	if(nFlag){
		if(nKey[0]){
			k_scan = nKey[0];
			printk("k_scan = %d\n", k_scan);
		}
	}
	
	//this delay can't more time
	delay(10000);
	HwADCCFG |= (1<<3); // Enable ADC IRQ
	HwADCCON &= ~(1<<4);
	
	return IRQ_HANDLED;
}

/**************************************************
 *	btn device functions
 *
 *************************************************/
static ssize_t btn_read(struct file *filp, char *buffer, size_t count, loff_t *ppos)
{
	if(k_scan){
		*buffer = (char)k_scan;
		k_scan = 0;
		return 1;
	}
	else
		return 0;
}

static ssize_t btn_write(struct file *filp, const char *buffer, size_t count, loff_t *ppos)
{
	return 0;
}

static unsigned int btn_poll(struct file *filp, struct poll_table_struct *wait)
{
	return 0; 
}

static int btn_open(struct inode *inode, struct file *filp)
{
	enable_irq(IRQ_ADC);
	return 0;
}

static int btn_release(struct inode *inode, struct file *filp)
{
	disable_irq(IRQ_ADC);
	return 0;
}

static struct file_operations tcc7901_fops = {
	owner:	THIS_MODULE,
	open:	btn_open,
	read:	btn_read,	
	write: btn_write,
	release:	btn_release,
	poll:	btn_poll,
};

static int __init btn_probe(struct device *dev)
{
	int ret;
	
	printk(DEVICE_NAME " initialized\n");
	
	ret = register_chrdev(200, DEVICE_NAME, &tcc7901_fops);
	if (ret < 0) {
		printk(DEVICE_NAME " can't get major number\n");
		return ret;
	}
	BtnMajor = ret;
	printk("%s device driver MAJOR:%d\n", DEVICE_NAME, BtnMajor);
	
	//ADC init
	HwADCCON  = (1<<4);
	HwPCK_ADC = (1<<26) | (1<<28) | 1;
	HwADCCFG  = (7<<8) | (1<<1) | 1;
	
	// Enable BTN interrupt
	//HwSEL |= (1<<30);
	//HwIRQMSK |= (1<<30);
	ret = request_irq(IRQ_ADC, btn_handler, IRQF_SHARED, 
			  DEVICE_NAME, btn_handler);
	if (ret)
		printk("BTN request IRQ failed\n");
	
	// Enable ADC
	HwADCCFG  |= (1<<3);
	HwADCCON &= ~(1<<4);	 
	
	return ret;
}

static struct device_driver btn_driver = {
	.name		= DEVICE_NAME,
	.probe		= btn_probe,
};

static int __init btn_init(void)
{
	return btn_probe(NULL);
}


static void __exit btn_exit(void)
{
	unregister_chrdev(BtnMajor, DEVICE_NAME);
	
	free_irq(IRQ_ADC, btn_handler);
	driver_unregister(&btn_driver);
}

module_init(btn_init);
module_exit(btn_exit);
