/*
 * tcc79x_ts.c
 *
 * touchScreen driver for tsc2003
 *
 * Updated @ 2007-06-13: Mac Chen <604sworder@163.com>
 *
 * Based on S3C2410-ts.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 *
 * History:
 *
 * 2008-04-14: migrate to TCC7901
 * 2007-04-17: V1.0 release
 * 
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/miscdevice.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>

#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/arch/tcc79x.h>

#ifdef CONFIG_I2C
#include <linux/i2c.h>
#else
#include "tcc7901_i2c.h"
#endif

/* debug macros */
#define DEBUG
#undef DEBUG
#ifdef DEBUG
#define DPRINTK( x... )	printk("%s: ", __FUNCTION__,  ##x)
#else
#define DPRINTK( x... )
#endif

#undef USE_ASYNC	
#define DEVICE_NAME	"tsc2003ts"
#define IRQ_TOUCH		2
//==============================================================
#define CMD_MEASURE_X		(12<<4)
#define CMD_MEASURE_Y		(13<<4)

#define CMD_XY_OFF			(0<<4)

#define CMD_POWER_0		(0<<2)
#define CMD_POWER_1		(1<<2)
#define CMD_POWER_2		(2<<2)
#define CMD_POWER_3		(3<<2)

#define READ_X	0
#define READ_Y	1

//unsigned short coor_x[1024], coor_y[1024];
typedef struct {
	unsigned short pressure;
	unsigned short x;
	unsigned short y;
	unsigned short pad;
} TS_RET;

#define PEN_UP	        0		
#define PEN_DOWN	1
#define PEN_FLEETING	2
#define MAX_TS_BUF	16	/* how many do we want to buffer */

typedef struct {
	unsigned int penStatus;		/* PEN_UP, PEN_DOWN, PEN_SAMPLE */
	TS_RET buf[MAX_TS_BUF];		/* protect against overrun */
	unsigned int head, tail;	/* head and tail for queued events */
	wait_queue_head_t wq;
	spinlock_t lock;
#ifdef USE_ASYNC
	struct fasync_struct *aq;
#endif
} TS_DEV;
static TS_DEV tsdev;

#define BUF_HEAD	(tsdev.buf[tsdev.head])
#define BUF_TAIL	(tsdev.buf[tsdev.tail])
#define INCBUF(x,mod) 	((++(x)) & ((mod) - 1))

#define TS_TIMER_DELAY  (HZ/100*6) /* 60 ms */
static struct timer_list ts_timer;
#define TS_PENUP_DELAY	8
static struct timer_list tmr_penup;
static TS_RET curPoint;
static unsigned long n_jiffies;

static int tsMajor = 0;

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/
//#define TSC2003_SLAVE_ADDRESS 0x90
#define TSC2003_SLAVE_ADDRESS 0x48
static unsigned short normal_i2c[] = {TSC2003_SLAVE_ADDRESS, I2C_CLIENT_END};
static unsigned short dummy[] = {I2C_CLIENT_END};

static struct i2c_client_address_data addr_data = {
	.normal_i2c = normal_i2c,
	.probe = dummy,
	.ignore = dummy,
};

static struct i2c_driver tsc2003_i2c_driver;
static struct i2c_client *tsc2003_i2c_client;

static int tsc2003_i2c_probe(struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *i2c;

	DPRINTK("\n");

	i2c = kzalloc(sizeof(*i2c), GFP_KERNEL);
	if (!i2c)
		return -ENOMEM;

	strcpy(i2c->name, "TSC2003");
	i2c->flags = 0;
	i2c->addr = addr;
	i2c->adapter = adap;
	i2c->driver = &tsc2003_i2c_driver;

	tsc2003_i2c_client = i2c;

	return i2c_attach_client(i2c);
}

static int tsc2003_i2c_detach(struct i2c_client *client)
{
	DPRINTK("\n");
	i2c_detach_client(client);
	kfree(tsc2003_i2c_client);
	return 0;
}

static int tsc2003_i2c_attach(struct i2c_adapter *adap)
{
	int ret;
	DPRINTK("\n");
	if (adap->nr == 1)	/* TSC2003 use the I2C Channel 1 */
		ret =  i2c_probe(adap, &addr_data, tsc2003_i2c_probe);
	else				/* If I2C Channel 1, return error */
		ret = -1;
	return ret;
}

static struct i2c_driver tsc2003_i2c_driver = {
	.driver = {
		.name = "TSC2003",
		.owner = THIS_MODULE,
	},
	.id = I2C_DRIVERID_TSC2003,
	.attach_adapter = tsc2003_i2c_attach,
	.detach_client = tsc2003_i2c_detach,
};

/*--------------------------------------------------------------*/
/*--------------------------------------------------------------*/
static void write_cmd_2003(unsigned char cmd)
{
#ifdef CONFIG_I2C
	i2c_master_send(tsc2003_i2c_client, &cmd, 1);
#else
	TOUCH_SEND_CMD(0, (unsigned int)(&cmd));
	udelay(1);
#endif
}

static short read_2003(int channel)
{
	unsigned char data[]={0,0};
	short value;
	
	if(channel == READ_X)
		write_cmd_2003(CMD_MEASURE_X | CMD_POWER_1);
	else if(channel == READ_Y)
		write_cmd_2003(CMD_MEASURE_Y | CMD_POWER_1);
	
#ifdef CONFIG_I2C
	i2c_master_recv(tsc2003_i2c_client, &data[0], 2);
#else
	TOUCH_READ_DAT(0, (unsigned int)(&data[0]));
#endif
	
	value = (data[0] & 0x00ff) << 4;
	value |= ((data[1] & 0x00ff) >> 4);
	value &= 0xfff;
	
	udelay(1);
	return(value>>1);
}

static void enable_touch_irq(void)
{
	//write_cmd_2003(CMD_XY_OFF | CMD_POWER_0);
	HwIRQMSK |= HwINT_EI2;
	HwIEN |= HwINT_EI2;
}

static void disable_touch_irq(void)
{
	//write_cmd_2003(CMD_XY_OFF | CMD_POWER_1);
	HwIRQMSK &= ~HwINT_EI2;
	
	HwIEN &= ~HwINT_EI2;
	HwCLR = HwINT_EI2;

}

//for the digi_int_handle
#define WAVE_LIMIT		5
#define START_AREA		4
#define COLLECT_NR		16 //every sampled must more than 10 dots
#define VALID_VALUE		10
static int get_raw_data(short * x, short * y)
{
	unsigned char  i, j, k, start, end, number;
	unsigned int dx, dy;
	unsigned short r_x[COLLECT_NR], r_y[COLLECT_NR];
	unsigned long  x_tol, y_tol;
	unsigned short m_pos_x, m_pos_y;
	
	* x = * y = 0;
	m_pos_x = m_pos_y = 0;
	
	//A/D conversion channel X
	j = 0;
	r_x[j] = read_2003(READ_X);
	udelay(5);
	r_x[j] = read_2003(READ_X);
	for(i = 0; i < COLLECT_NR; i ++){
		r_x[j] = read_2003(READ_X);
		if(r_x[j] != 0)
		{
			//printk("(%x,?),",r_x[j]);
			j ++;
		}
	}
	//printk("\n");
	
	//A/D conversion channel Y
	k = 0;
	r_y[k] = read_2003(READ_Y);
	udelay(5);
	r_y[k] = read_2003(READ_Y);
	for(i = 0; i < COLLECT_NR; i ++){
		r_y[k] = read_2003(READ_Y);
		if(r_y[k] != 0) 
		{
			//printk("(?,%x),",r_y[k]);
			k ++;
		}
	}
	//printk("\n");
	
	//Hardware PENUP detect
	write_cmd_2003(CMD_XY_OFF | CMD_POWER_0);
	udelay(100);
	if(HwGPADAT & (1<<2)){
		//printk("Hardware PENUP\n");
		return -1;
	}
	
	j = (j<k)?j:k;
	
	if(j <= VALID_VALUE){
		//printk("error 1\n");
		goto digi_int_exit;
	}
	
	//find the first effective result among first START_AREA results
	//discard the first 2 results
	for(i = 2; i < START_AREA; i ++){
		dx = (r_x[i] > r_x[i+1]) ? (r_x[i] - r_x[i+1]):(r_x[i+1] - r_x[i]);
		dx = dx * 100 / r_x[i+1];
		
		dy = (r_y[i] > r_y[i+1]) ? (r_y[i] - r_y[i+1]):(r_y[i+1] - r_y[i]);
		dy = dy * 100 / r_y[i+1];
		
		if(dx < WAVE_LIMIT && dy < WAVE_LIMIT)
			break;
	}
	//if not find start result
	if((start = i) >= START_AREA){
		//printk("error 2\n");
		goto digi_int_exit;
	}
	
	//printk("start = %d\n",start);
	for(i = start + 1; i < j - 1; i ++){
		dx = (r_x[i] > r_x[i+1]) ? (r_x[i] - r_x[i+1]):(r_x[i+1] - r_x[i]);
		dx = dx * 100 / r_x[i+1];
		
		dy = (r_y[i] > r_y[i+1]) ? (r_y[i] - r_y[i+1]):(r_y[i+1] - r_y[i]);
		dy = dy * 100 / r_y[i+1];
		
		if(dx > WAVE_LIMIT || dy > WAVE_LIMIT)
			break;
	}
	end = i;
	
	// valid value is too less to continue
	if(end - start < VALID_VALUE){
		//printk("error 3\n");
		goto digi_int_exit;
	}
	
	x_tol = y_tol = 0; //sum the coordinate values
	number = 1 + end - start;
	for(i = start; i < end + 1; i++){
		x_tol += r_x[i];
		y_tol += r_y[i];
	}
	
	if(x_tol < 0x10000) //calculate coordinate_x 's average
		m_pos_x = (unsigned short)x_tol / number;
	else if(x_tol >= 0x10000 && x_tol < 0x20000)
		m_pos_x = (((unsigned short)(x_tol >> 1)) / number) << 1;
	else if(x_tol >= 0x20000 && x_tol < 0x40000)
		m_pos_x = (((unsigned short)(x_tol >> 2)) / number) << 2;
	else if(x_tol >= 0x40000 && x_tol < 0x80000)
		m_pos_x = (((unsigned short)(x_tol >> 3)) / number) << 3;
	else if(x_tol >= 0x80000 && x_tol < 0x100000)
		m_pos_x = (((unsigned short)(x_tol >> 4)) / number) << 4;
	else{
		//printk("error 4\n");
		goto digi_int_exit;
	}
	
	if(y_tol < 0x10000) //claculate coordinate_y 's average
		m_pos_y = (unsigned short)y_tol / number;
	else if(y_tol >= 0x10000 && y_tol < 0x20000)
		m_pos_y = (((unsigned short)(y_tol >> 1)) / number) << 1;
	else if(y_tol >= 0x20000 && y_tol < 0x40000)
		m_pos_y = (((unsigned short)(y_tol >> 2)) / number) << 2;
	else if(y_tol >= 0x40000 && y_tol < 0x80000)
		m_pos_y = (((unsigned short)(y_tol >> 3)) / number) << 3;
	else if(y_tol >= 0x80000 && y_tol < 0x100000)
		m_pos_y = (((unsigned short)(y_tol >> 4)) / number) << 4;
	else{
		//printk("error 5\n");
		goto digi_int_exit;
	}
	

digi_int_exit:
	* x = m_pos_x;
	* y = m_pos_y;
	
	if(m_pos_x && m_pos_y)
		return 0;
	else
		return -1;
}

//int point, read_point;
static irqreturn_t tsc2003_handler(int irq, void *dev_id)
{
	unsigned short t_x, t_y;
	
	//printk("touch_interrupt\n");
	spin_lock_irq(&(tsdev.lock));
	if(!get_raw_data(&t_x, &t_y))
	{
		//printk("Interrupt,x=%4d,y=%4d\n", t_x, t_y);
		BUF_HEAD.x = t_x/2;
		BUF_HEAD.y = t_y/2;
		BUF_HEAD.pressure = PEN_DOWN;
		printk("PEN DOWN:%d,%d\n", BUF_HEAD.x, BUF_HEAD.y);
		
		curPoint.x = BUF_HEAD.x;
		curPoint.y = BUF_HEAD.y;
		curPoint.pressure = BUF_HEAD.pressure;
		n_jiffies = jiffies;
		
		del_timer(&tmr_penup);
		tmr_penup.expires = jiffies + TS_PENUP_DELAY;
		add_timer(&tmr_penup);
		
		tsdev.head = INCBUF(tsdev.head, MAX_TS_BUF);
				
		wake_up_interruptible(&(tsdev.wq));
	}
	
	write_cmd_2003(CMD_XY_OFF | CMD_POWER_0);
	//delay(2000); //some delay should after power off command, otherwise interrupt will error
	
	spin_unlock_irq(&(tsdev.lock));
	disable_touch_irq();
	
	return IRQ_HANDLED;
}

// use a timer to fire PENUP event
static void tsc2003_penup(unsigned long tmp)
{
	del_timer(&tmr_penup);
	
	BUF_HEAD.x = curPoint.x;
	BUF_HEAD.y = curPoint.y;
	BUF_HEAD.pressure = PEN_UP;
	
	tsdev.head = INCBUF(tsdev.head, MAX_TS_BUF);
	
	printk("PENUP\n");
}

static ssize_t tsc2003_ts_read(struct file *filp, char *buffer, size_t count, loff_t *ppos)
{
	TS_RET ts_ret;
	
retry: 
	if (tsdev.head != tsdev.tail)
	{
		spin_lock_irq(&(tsdev.lock));
		//printk("ts_read, (%d, %d)\n", ts_ret.x, ts_ret.y);
		ts_ret.x = BUF_TAIL.x;
		ts_ret.y = BUF_TAIL.y;
		ts_ret.pressure = BUF_TAIL.pressure;
		copy_to_user(buffer, (char *)&ts_ret, sizeof(TS_RET));
		tsdev.tail = INCBUF(tsdev.tail, MAX_TS_BUF);
		spin_unlock_irq(&(tsdev.lock));
		
		return(sizeof(TS_RET));
	}
	else
	{
	/*
		printk("jiffies:%lu, %lu,\n", jiffies, n_jiffies);
		if ((jiffies - n_jiffies > TS_PENUP_DELAY) && (curPoint.x != 0) )
		//if (curPoint.x != 0)
		{
			curPoint.pressure = PEN_UP;
			copy_to_user(buffer, (char *)&curPoint, sizeof(TS_RET));
			return (sizeof(TS_RET));
		}
	*/	
		if (filp->f_flags & O_NONBLOCK)  // necessary, otherwise the app will be blocked
			return -EAGAIN;
		interruptible_sleep_on(&(tsdev.wq));
		if (signal_pending(current))
			return -ERESTARTSYS;
		goto retry;
	}

	return 0;
}

#ifdef USE_ASYNC
static int tsc2003_ts_fasync(int fd, struct file *filp, int mode) 
{
	return fasync_helper(fd, filp, mode, &(tsdev.aq));
}
#endif

static unsigned int tsc2003_ts_poll(struct file *filp, struct poll_table_struct *wait)
{
	//printk("tsc2003_ts_poll\n");
	poll_wait(filp, &(tsdev.wq), wait);
	if (tsdev.head == tsdev.tail)
		return 0;
	else
		return  (POLLIN | POLLRDNORM);
}

static int tsc2003_ts_open(struct inode *inode, struct file *filp)
{
	//printk("Touch screen device open.\n");
	
	return 0;
}

static int tsc2003_ts_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static struct file_operations tsc2003_fops = {
	owner:	THIS_MODULE,
	open:	tsc2003_ts_open,
	read:	tsc2003_ts_read,	
	release:	tsc2003_ts_release,
#ifdef USE_ASYNC
	fasync:	tsc2003_ts_fasync,
#endif
	poll:	tsc2003_ts_poll,
};


static void ts_timer_handler(unsigned long data)
{
	static int flag=0;
	
	spin_lock_irq(&(tsdev.lock));
	del_timer(&ts_timer);
	if (flag)
		enable_touch_irq();
	else
		disable_touch_irq();
	
	ts_timer.expires = jiffies + TS_TIMER_DELAY;
	add_timer(&ts_timer);
	flag = !flag;
			
	spin_unlock_irq(&(tsdev.lock));
}

static int __init tsc2003ts_probe(struct device *dev)
{
	int ret;

	//printk("tsc2003ts_probe\n");
	
	ret = register_chrdev(0, DEVICE_NAME, &tsc2003_fops);
	if (ret < 0) {
	  printk(DEVICE_NAME " can't get major number\n");
	  return ret;
	}
	tsMajor = ret;
	printk("%s device driver MAJOR:%d\n", DEVICE_NAME, tsMajor);

#ifdef CONFIG_I2C
	ret = i2c_add_driver(&tsc2003_i2c_driver);
	if (ret != 0)
		printk("%s: can't add i2c driver\n", __func__);
#else
	TOUCH_IIC_INIT();
#endif
	
	//EINT_2 init
	HwPORTCFG4 &= ~(0x1f<<16);
	HwGPAEN &= ~(1<<2);
	//HwCPUD3 = (HwCPUD3 & (~(0x3<<0)) | (1<<1);
	HwEINTSEL0 = (HwEINTSEL0 & (~(0x1f<<16))) | (18<<16);
	
	HwSEL |= HwINT_EI2;
	HwIRQMSK |= HwINT_EI2;
	HwIEN &= ~HwINT_EI2;
	
	//set IRQ as level sensitive; low active
	HwPOL |= HwINT_EI2;
	HwMODE |= HwINT_EI2;
	
	write_cmd_2003(CMD_XY_OFF | CMD_POWER_0);
	
	/* Enable touch interrupt */
	ret = request_irq(IRQ_TOUCH, tsc2003_handler, IRQF_SHARED, 
			  DEVICE_NAME, tsc2003_handler);
	if (ret)
		return ret;
	enable_touch_irq();
	
	/* Wait for touch screen interrupts */
	init_timer(&ts_timer);
	ts_timer.function = ts_timer_handler;
	ts_timer.expires = jiffies + TS_TIMER_DELAY;
	add_timer(&ts_timer);

	// PENUP timer
	init_timer(&tmr_penup);
	tmr_penup.function = tsc2003_penup;
	tmr_penup.expires = jiffies + TS_PENUP_DELAY;
	add_timer(&tmr_penup);
		
	init_waitqueue_head(&(tsdev.wq));
	tsdev.head = tsdev.tail = 0;
	tsdev.penStatus = PEN_UP;
	
#ifdef CONFIG_DEVFS_FS
	devfs_mk_dir("touchscreen");
	devfs_mk_cdev(MKDEV(tsMajor, TSRAW_MINOR), S_IFCHR|S_IRUGO|S_IWUSR, "touchscreen/%d", 0);
#endif

	printk(DEVICE_NAME " initialized\n");

	return 0;
}

#ifdef CONFIG_PM
static int tsc2003ts_suspend(struct device *pdev, pm_message_t state)
{
    return 0;
}

static int tsc2003ts_resume(struct device *pdev)
{
    return 0;
}
#endif

static struct device_driver tsc2003ts_driver = {
	.name		= DEVICE_NAME,
	.probe		= tsc2003ts_probe,
#ifdef CONFIG_PM
	.suspend	= tsc2003ts_suspend,
	.resume		= tsc2003ts_resume,
#endif
};

static int __init tsc2003ts_init(void)
{
	printk("TCC7901 touch panel driver init\n");
	return (tsc2003ts_probe(NULL));
	/*
	ret = driver_register(&tsc2003ts_driver);
	if(ret)
		printk("register %s driver failed, return code is %d\n", DEVICE_NAME, ret);
	return ret;
	*/
}


static void __exit tsc2003ts_exit(void)
{
#ifdef CONFIG_DEVFS_FS	
	devfs_remove("touchscreen/%d", 0);
	devfs_remove("touchscreen");
#endif	
	unregister_chrdev(tsMajor, DEVICE_NAME);
	
	free_irq(IRQ_TOUCH, tsc2003_handler);
	driver_unregister(&tsc2003ts_driver);
}

module_init(tsc2003ts_init);
module_exit(tsc2003ts_exit);

