/*
 * File:        drivers/char/tcc_ioctl.c
 *
 * Created:     June 10, 2008
 * Copyright (C) 2008-2009 Telechips <linux@telechips.com>
 * Description: User-level Clock Control IOCTL driver
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/ioctl.h>

#include <asm/plat-tcc/tcc_ckc.h>
#include <linux/tcc_ioctl.h>

#ifdef CONFIG_ARCH_TCC79X
#include <asm/arch/tcc79x.h>
#elif CONFIG_ARCH_TCC83X
#include <asm/arch/tcc83x_virt.h>
#endif

#ifndef ON
#define ON 1
#endif

#ifndef OFF
#define OFF    0
#endif

#define DEV_NAME "tcc_ioctl"
#define BUF_LEN 80

static int tcc_open(struct inode *inode, struct file *file)
{
    return 0;
}

static int tcc_release(struct inode *inode, struct file *file)
{
    return 0;
}

int tcc_ioctl(struct inode *inode, struct file *file,
                 unsigned int cmd, unsigned long arg)
{
    struct ckc_ioctl st;

    memset(&st, 0, sizeof(struct ckc_ioctl));

    switch(cmd)
    {
    case IOCTL_CKC_SET_PERICLOCK:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        tca_ckc_setperi( st.in_ckc.PeriClockName, st.in_ckc.PeriUnit,
                          st.in_ckc.PeriFreq, st.in_ckc.PeriClockSource);
        break;

    case IOCTL_CKC_GET_PERICLOCK:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        st.out_ckc.PeriFreq = tca_ckc_getperi( st.in_ckc.PeriClockName);
        if(copy_to_user((void *)arg, &st, sizeof(struct ckc_ioctl)))
            return -EFAULT;
        break;

    case IOCTL_CKC_SET_PERIBUSCLKCTL:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        tca_ckc_setperibus(st.in_ckc.PeriRBName, st.in_ckc.Mode); // Enable or eisable
        break;

    case IOCTL_CKC_SET_PERISWRESET:
        if(copy_from_user((void *)arg, &st, sizeof(struct ckc_ioctl)))
            return -EFAULT;
        tca_ckc_resetperi(st.in_ckc.PeriRBName, ON);
        tca_ckc_resetperi(st.in_ckc.PeriRBName, OFF);
        break;

    case IOCTL_CKC_SET_MAINCLOCKCFG:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        st.out_ckc.retVal = tca_ckc_setclock(st.in_ckc.PllValue,st.in_ckc.CpuValue,
                                        st.in_ckc.BusValue,st.in_ckc.PllChannel);
        st.out_ckc.CurrentSysFreq = tca_ckc_getcurrentsys();
        st.out_ckc.CurrentBusFreq = tca_ckc_getcurrentbus();
        if(copy_to_user((void *)arg, &st, sizeof(struct ckc_ioctl)))
            return -EFAULT;
        break;

    case IOCTL_CKC_GET_PMSCLOCKCFG:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        tca_ckc_setpll(st.in_ckc.P,st.in_ckc.M,st.in_ckc.S,st.in_ckc.PllChannel);
        st.out_ckc.CurrentSysFreq = tca_ckc_getcurrentsys();
        st.out_ckc.CurrentBusFreq = tca_ckc_getcurrentbus();
        if(copy_to_user((void *)arg, &st, sizeof(struct ckc_ioctl)))
            return -EFAULT;
        break;

    case IOCTL_CKC_GET_CLOCKINFO:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        st.out_ckc.CurrentSysFreq = tca_ckc_currentpll(st.in_ckc.PllChannel);//CKC_GetCurrentSysClock();
        st.out_ckc.CurrentCpuFreq = tca_ckc_getcurrentcpu();
        st.out_ckc.CurrentBusFreq = tca_ckc_getcurrentbus();
        if(copy_to_user((void *)arg, &st, sizeof(struct ckc_ioctl)))
            return -EFAULT;
        break;
    case IOCTL_CKC_GET_VALIDPLLINFO:
        if(copy_from_user(&st, (void *)arg, sizeof(struct ckc_ioctl)))
           return -EFAULT;
        tca_ckc_validpll(st.out_ckc.ValidPll);
        break;

    default :
        break;
    }

    return 0;
}

struct file_operations tcc_fops = {
    .owner      = THIS_MODULE,
    .ioctl      = tcc_ioctl,
    .open       = tcc_open,
    .release    = tcc_release,
};


int init_module(void)
{
    int ret;

    ret = register_chrdev(MAJOR_NUM, DEV_NAME, &tcc_fops);

    if(ret < 0){
        printk(KERN_ALERT " %s failed with %d\n", "fail to register the character device", ret);
        return ret;
    }

    return 0;
}

void cleanup_module(void)
{
    unregister_chrdev(MAJOR_NUM, DEV_NAME);
}


module_init(init_module);
module_exit(cleanup_module);


MODULE_AUTHOR("Telechips Inc. SYS4-3 linux@telechips.com");
MODULE_DESCRIPTION("TCC ioctl driver");
MODULE_LICENSE("GPL");


