/*
 * linux/drivers/video/tcc_phisical_mem.c
 *
 * Author:  <linux@telechips.com>
 * Created: November 12, 2010
 * Description: TCC Physical Memory Allocation Driver
 *
 * Copyright (C) 20010-2011 Telechips 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see the file COPYING, or write
 * to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/highmem.h>
#ifdef CONFIG_PM
#include <linux/pm.h>
#endif
#ifdef CONFIG_HAS_EARLYSUSPEND
#include <linux/earlysuspend.h>
#endif
#include <mach/tcc_used_mem.h>
#include <mach/tcc_physical_mem_ioctl.h>

/*****************************************************************************

 VARIABLES

******************************************************************************/

/* Debugging stuff */
static int debug = 0;
#define dprintk(msg...)	if (debug) { printk( "tcc_physical_mem: " msg); }

#define MEM_DEVICE_NAME	"tcc_mem"
#define MAJOR_ID		207
#define MINOR_ID		1


static struct class *tcc_physical_mem_class;


/*****************************************************************************

 FUNCTIONS

******************************************************************************/

/*****************************************************************************
 Function Name : tcc_physical_mem_ioctl()
******************************************************************************/
static int tcc_physical_mem_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	stTCC_PHYSICAL_MEM_TYPE physical_mem;
	int ret_val = 0;
 	
	dprintk("tcc_physical_mem_ioctl IOCTRL[%d]\n", cmd);

	switch(cmd)
	{
		case MEM_TYPE_PMEM_SURF_BASE:
			physical_mem.mem_addr = PMEM_SURF_BASE;
			physical_mem.mem_size = PMEM_SURF_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;

		case MEM_TYPE_PMEM_VIQE_BASE:
			physical_mem.mem_addr = (PMEM_SURF_BASE + PMEM_SURF_SIZE);
			physical_mem.mem_size = PMEM_VIQE_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;
			
		case MEM_TYPE_PMEM_CAM_BASE:
			physical_mem.mem_addr = (PMEM_VIQE_BASE + PMEM_VIQE_SIZE);
			physical_mem.mem_size = PMEM_CAM_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;
			
		case MEM_TYPE_OVERLAY_MEM_BASE:
			physical_mem.mem_addr = (PMEM_CAM_BASE + PMEM_CAM_SIZE);
			physical_mem.mem_size = OVERLAY_MEM_TOTAL_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;

		case MEM_TYPE_VIDEO_MEM_BASE:
			physical_mem.mem_addr = (OVERLAY_MEM_BASE + OVERLAY_MEM_TOTAL_SIZE);
			physical_mem.mem_size = VIDEO_MEM_TOTAL_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;
			
		case MEM_TYPE_COPYBIT_MEM_BASE:
			physical_mem.mem_addr = (VIDEO_MEM_BASE + VIDEO_MEM_TOTAL_SIZE);
			physical_mem.mem_size = COPYBIT_MEM_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;
			
		case MEM_TYPE_FB_VIDEO_MEM_BASE:
			physical_mem.mem_addr = (COPYBIT_MEM_BASE + COPYBIT_MEM_SIZE);
			physical_mem.mem_size = FB_VIDEO_MEM_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;

		case MEM_TYPE_FB_MEM_BASE:
			physical_mem.mem_addr = (FB_VIDEO_MEM_BASE + FB_VIDEO_MEM_SIZE);
			physical_mem.mem_size = FB_MEM_TOTAL_SIZE;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;
			
		case MEM_TYPE_STREAM_MEM_BASE:
			#ifdef STREAM_MEM_SIZE
			physical_mem.mem_addr = (FB_MEM_BASE + FB_MEM_TOTAL_SIZE);
			physical_mem.mem_size = STREAM_MEM_SIZE;
			#else
			physical_mem.mem_addr = 0;
			physical_mem.mem_size = 0;
			#endif
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;

		default:
			physical_mem.mem_addr = 0;
			physical_mem.mem_size = 0;
			if (copy_to_user((unsigned int*)arg, &physical_mem, sizeof(stTCC_PHYSICAL_MEM_TYPE)))
				ret_val = -EFAULT;
			else
				ret_val = 0;
			break;
			printk(" Unsupported IOCTL!!!\n");
			break;			
	}

	dprintk("tcc_physical_mem_ioctl : mem_addr[0x%08x] mdm_size[0x%08x]\n", physical_mem.mem_addr, physical_mem.mem_size);

	return ret_val;
}

/*****************************************************************************
 Function Name : tcc_physical_mem_open()
******************************************************************************/
static int tcc_physical_mem_open(struct inode *inode, struct file *file)
{	
	dprintk("%s  \n", __func__);

	return 0;
}

/*****************************************************************************
 Function Name : tcc_physical_mem_release()
******************************************************************************/
static int tcc_physical_mem_release(struct inode *inode, struct file *file)
{
	dprintk(" %s  \n", __func__);

	return 0;
}


static struct file_operations tcc_physical_mem_fops = 
{
	.owner			= THIS_MODULE,
	.ioctl			= tcc_physical_mem_ioctl,
	.open			= tcc_physical_mem_open,
	.release		= tcc_physical_mem_release,	
};


/*****************************************************************************
 Function Name : tcc_physical_mem_init()
******************************************************************************/
int __init tcc_physical_mem_init(void)
{
	printk("%s  \n", __func__);

	register_chrdev(MAJOR_ID, MEM_DEVICE_NAME, &tcc_physical_mem_fops);
	tcc_physical_mem_class = class_create(THIS_MODULE, MEM_DEVICE_NAME);
	device_create(tcc_physical_mem_class, NULL, MKDEV(MAJOR_ID, MINOR_ID), NULL, MEM_DEVICE_NAME);
	
	return 0;
}

/*****************************************************************************
 Function Name : tcc_physical_mem_exit()
******************************************************************************/
void __exit tcc_physical_mem_exit(void)
{
	printk("%s  \n", __func__);

	unregister_chrdev(MAJOR_ID, MEM_DEVICE_NAME);

	return;
}

module_init(tcc_physical_mem_init);
module_exit(tcc_physical_mem_exit);

MODULE_AUTHOR("Telechps");
MODULE_DESCRIPTION("Telechips Physical Memory Allocation driver");
MODULE_LICENSE("GPL");


