/*
 * drivers/char/tcc_vpu.c
 *
 * TCC VPU driver
 *
 * Copyright (C) 2009 Telechips, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/mman.h>
#include <linux/init.h>
#include <linux/ptrace.h>
#include <linux/device.h>
#include <linux/highmem.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/syscalls.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <mach/globals.h>
#include <mach/tca_ckc.h>
#include <mach/bsp.h>

#include <mach/tcc_vpu_ioctl.h>
#include <linux/delay.h>
#include <linux/cpufreq.h>

#if 0
static int debug	   = 1;
#else
static int debug	   = 0;
#endif

static OPENED_gINFO opened_info;

#define dprintk(msg...)	if (debug) { printk( "TCC_VPU: " msg); }

#define DEV_MAJOR	196
#define DEV_MINOR	1
#define DEV_NAME	"vpu"

static struct clk *vcore_clk;
static struct clk *vcache_clk; // for pwdn and vBus.
static struct clk *vcodec_clk; // for pwdn and vBus.
static struct clk *overlay_clk;

static int curr_type = 0;

#if defined(CONFIG_DRAM_MDDR)
	#if defined(CONFIG_HIGH_PERFORMANCE)
		#define CPU_1ST_CLOCK 600000
		#define CPU_2ND_CLOCK 500000
		#define CPU_3RD_CLOCK 425250
		
		#define MEM_1ST_CLOCK 290000
		#define MEM_2ND_CLOCK 240000
		#define MEM_3RD_CLOCK 198000
	#else
		#define CPU_1ST_CLOCK 500000
		#define CPU_2ND_CLOCK 425250
		#define CPU_3RD_CLOCK 364500
		
		#define MEM_1ST_CLOCK 240000
		#define MEM_2ND_CLOCK 198000
		#define MEM_3RD_CLOCK 166000
	#endif
#elif defined(CONFIG_DRAM_DDR2) || defined(CONFIG_DRAM_DDR3)
	#if defined(CONFIG_TCC_CLK_720_360) || defined(CONFIG_TCC_CLK_720_330)
		#define CPU_1ST_CLOCK 720000
		#define CPU_2ND_CLOCK 600000
		#define CPU_3RD_CLOCK 425250
		
		#define MEM_1ST_CLOCK 330000
		#define MEM_2ND_CLOCK 290000
		#define MEM_3RD_CLOCK 198000
	#elif defined(CONFIG_TCC_CLK_600_300)
		#define CPU_1ST_CLOCK 600000
		#define CPU_2ND_CLOCK 500000
		#define CPU_3RD_CLOCK 425250
		
		#define MEM_1ST_CLOCK 290000
		#define MEM_2ND_CLOCK 240000
		#define MEM_3RD_CLOCK 198000
	#else
		#define CPU_1ST_CLOCK 500000
		#define CPU_2ND_CLOCK 425250
		#define CPU_3RD_CLOCK 364500
		
		#define MEM_1ST_CLOCK 240000
		#define MEM_2ND_CLOCK 198000
		#define MEM_3RD_CLOCK 166000
	#endif
#endif

static unsigned int gtVpuSizeTable[] =
{
//	176 * 144 , //QCIF,	
//	352 * 288 , //CIF,	
//	400 * 240 , //WQVGA,
	
//	640 * 480 , //VGA,	
	720 * 480 , //D1,	
//	800 * 600 , //SVGA,	
	1280 * 720, //HD720P,
#if defined(CONFIG_MACH_TCC8900) || defined(CONFIG_ARCH_TCC93XX)
	1920 * 1080, //HD1080P,
#endif
};


#define SIZE_TABLE_MAX 				(sizeof(gtVpuSizeTable)/sizeof(unsigned int))

//LCD :: portrait WVGA 
extern struct tcc_freq_table_t gtVpuHighClockLimitTable[];

//LCD :: portrait HVGA or landscape WVGA 
extern struct tcc_freq_table_t gtVpuNormalClockLimitTable[];

static int Get_Index_sizeTable(unsigned int Image_Size)
{
	int i;
	
	//exception process in case of error-size in parser.!!
	if(Image_Size < 64*64)
		Image_Size = 1920*1080;

	for(i=0; i<SIZE_TABLE_MAX; i++)
	{
		if( gtVpuSizeTable[i] >= Image_Size)
		{
			return i;
		}
	}
	
	return (SIZE_TABLE_MAX -1);
}

#ifdef CONFIG_VIDEO_TCCXX_CAMERA
extern void cif_set_frameskip(unsigned char skip_count, unsigned char locked);
extern int tccxxx_isalive_camera(void);
#endif

extern int tccxxx_sync_player(int sync);
static int cache_droped = 0;
int vpu_setClock(unsigned int type, unsigned int img_width, unsigned int img_height, unsigned int bitrate, unsigned int framerate)
{
	int index, img_size;
	int lcd_width = SCREEN_WIDTH, lcd_height = SCREEN_HEIGHT;
	struct tcc_freq_table_t clk_table;

	img_size = img_width * img_height;
	index = Get_Index_sizeTable(img_size);
	if((index < (SIZE_TABLE_MAX-1)) && framerate > 30)
		index += 1;

	if(!cache_droped)
	{
		tccxxx_sync_player(1);
		cache_droped = 1;
	}

	//portrait-lcd and over HVGA lcd => Rotate + Scaler  :: This need high-performance!!
	if(type == CONTENT_VIDEO_SWCODEC)
	{
		dprintk("VPU CLOCK :: sw codec max-performance!! \n");
		index = SIZE_TABLE_MAX - 1;
		memcpy(&clk_table, &gtVpuHighClockLimitTable[index], sizeof(struct tcc_freq_table_t));
	}
	else
	{
		if(lcd_width < lcd_height && (lcd_width > 320 || lcd_height > 480)){
			dprintk("VPU CLOCK :: high-performance!! \n");
			memcpy(&clk_table, &gtVpuHighClockLimitTable[index], sizeof(struct tcc_freq_table_t));
		}
		else {//landscape-lcd or smaller lcd than HVGA => This need normal-performance!! 
			dprintk("VPU CLOCK :: normal-performance!! %d \n", index);
			memcpy(&clk_table, &gtVpuNormalClockLimitTable[index], sizeof(struct tcc_freq_table_t));
		}
	}
	
	printk("Clk_idx[%d] :: res = %d x %d, fps: %d, mbps: %d \n", index, img_width, img_height, framerate, bitrate);

	curr_type = type;

	if(type == CONTENT_VIDEO_SWCODEC)
		clk_table.cpu_freq = CPU_1ST_CLOCK;

	if(type != CONTENT_CAMCORDER)
	{
#if defined(CONFIG_MACH_TCC8900) || defined(CONFIG_ARCH_TCC93XX)
		if(bitrate > 25)
		{
			if(clk_table.cpu_freq < CPU_1ST_CLOCK)
				clk_table.cpu_freq = CPU_1ST_CLOCK;
			
			if(clk_table.mem_freq < MEM_1ST_CLOCK)
				clk_table.mem_freq = MEM_1ST_CLOCK;
		}
		else if(bitrate > 15)
		{
			if(clk_table.cpu_freq < CPU_2ND_CLOCK)
				clk_table.cpu_freq = CPU_2ND_CLOCK;
			
			if(clk_table.mem_freq < MEM_2ND_CLOCK)
				clk_table.mem_freq = MEM_2ND_CLOCK;
		}
		else if (bitrate > 10)
		{ 
			if(clk_table.cpu_freq < CPU_3RD_CLOCK)
				clk_table.cpu_freq = CPU_3RD_CLOCK;
			
			if(clk_table.mem_freq < MEM_3RD_CLOCK)
				clk_table.mem_freq = MEM_3RD_CLOCK;
		}
		else
		{
			/*NONE*/
		}
#else
		if(bitrate >= 7)
		{
			clk_table.cpu_freq = CPU_1ST_CLOCK;
			clk_table.mem_freq = MEM_1ST_CLOCK;
		}
		else if (bitrate >= 5)
		{
			clk_table.cpu_freq = CPU_2ND_CLOCK;
			clk_table.mem_freq = MEM_2ND_CLOCK;
		}
		else
		{
			/*NONE*/
		}
#endif
	}

	if(type == CONTENT_CAMCORDER)
	{
	#if defined(CONFIG_HIGH_PERFORMANCE)
		clk_table.vcod_freq = 1920000;
	#endif	
	
	#ifdef CONFIG_VIDEO_TCCXX_CAMERA
		cif_set_frameskip(5, 1); //Pause frame-update to prevent overflowing cif fifo buffer while memory clock is change. 
		tcc_cpufreq_set_limit_table(&clk_table, TCC_FREQ_LIMIT_VPU_ENC, 1);
		cif_set_frameskip(0, 0);
	#else
		tcc_cpufreq_set_limit_table(&clk_table, TCC_FREQ_LIMIT_VPU_ENC, 1);
	#endif
	}
	else
		tcc_cpufreq_set_limit_table(&clk_table, TCC_FREQ_LIMIT_VPU_DEC, 1);

	return 0;
}

int vpu_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	CONTENTS_INFO info;
	OPENED_sINFO open_info;
	
	switch(cmd)
	{
		case VPU_SET_CLK:
			if(copy_from_user(&info,(CONTENTS_INFO*)arg,sizeof(info)))
				return -EFAULT;
			return vpu_setClock(info.type, info.width, info.height, info.bitrate, info.framerate);

		case VPU_GET_OPEN_INFO:
			if (copy_to_user((unsigned int*)arg, &opened_info, sizeof(OPENED_gINFO)))
				return -EFAULT;
			else
				return 0;

		case VPU_SET_OPEN_INFO:
			if(copy_from_user(&open_info,(OPENED_sINFO*)arg,sizeof(OPENED_sINFO)))
				return -EFAULT;

			if(open_info.type == OPEN_VIDEO)
				opened_info.vid_opened = open_info.opened_cnt;
			else
				opened_info.dmb_opened = open_info.opened_cnt;
			return 0;
			
		default:
			printk(" Unsupported IOCTL!!!\n");      
			break;			
	}

	return 0;
}


static int vpu_open(struct inode *inode, struct file *filp)
{
	clk_enable(vcore_clk);
	clk_enable(vcache_clk);
	clk_enable(vcodec_clk);
	clk_enable(overlay_clk);

	opened_info.count++;
	dprintk("open :: vpu_opened %d \n", opened_info.count);

	return 0;
}

static unsigned int total_cnt = 0;
static int vpu_release(struct inode *inode, struct file *filp)
{
	PLCDC pLCDC;
	clk_disable(vcache_clk);
	clk_disable(vcodec_clk);
	clk_disable(vcore_clk);
	clk_disable(overlay_clk);

#ifdef CONFIG_VIDEO_TCCXX_CAMERA
	if(!tccxxx_isalive_camera())
#endif
	{	
		volatile PCKC pCKC;
		pCKC = (volatile PCKC)tcc_p2v(HwCLK_BASE);
		
		//Disable overlay-ch0 in vpu-close because of performance-dead.!!
#if defined(CONFIG_ARCH_TCC93XX)
		if(pCKC->PCLK_LCD1 & Hw29){
			pLCDC = (volatile PLCDC)tcc_p2v(HwLCDC1_BASE);
			BITCSET (pLCDC->LI0C, 0x1<<28, (0)<<28 | (1) << 16);
			mdelay(16);
		}
#else
		if(pCKC->PCLK_LCD1 & Hw28){
			pLCDC = (volatile PLCDC)tcc_p2v(HwLCDC1_BASE);
			BITCSET (pLCDC->LI0C, 0x1<<28, (0)<<28);		
			mdelay(16);
		}
#endif /* CONFIG_ARCH_TCC93XX */
#if defined(CONFIG_ARCH_TCC93XX)
		if(pCKC->PCLK_LCD0 & Hw29){
			pLCDC = (volatile PLCDC)tcc_p2v(HwLCDC0_BASE);
			BITCSET (pLCDC->LI0C, 0x1<<28, (0)<<28 | (1) << 16);
			mdelay(16);
		}
#endif /* CONFIG_ARCH_TCC93XX */
	}

	dprintk("VPU CLOCK ::rel initialized to 0 \n");

	if(curr_type == CONTENT_CAMCORDER)
		tcc_cpufreq_set_limit_table(&gtVpuNormalClockLimitTable[0], TCC_FREQ_LIMIT_VPU_ENC, 0);
	else
		tcc_cpufreq_set_limit_table(&gtVpuNormalClockLimitTable[0], TCC_FREQ_LIMIT_VPU_DEC, 0);

	opened_info.count--;
	total_cnt++;

	printk("release :: vpu_opened %d , total = %d \n", opened_info.count, total_cnt);

	return 0;
}

static void mmap_mem_open(struct vm_area_struct *vma)
{
	/* do nothing */
}

static void mmap_mem_close(struct vm_area_struct *vma)
{
	/* do nothing */
}

static struct vm_operations_struct vpu_mmap_ops = {
	.open = mmap_mem_open,
	.close = mmap_mem_close,
};

static inline int uncached_access(struct file *file, unsigned long addr)
{
	if (file->f_flags & O_SYNC)
		return 1;
	return addr >= __pa(high_memory);
}

static pgprot_t phys_mem_access_prot(struct file *file, unsigned long pfn,
				     unsigned long size, pgprot_t vma_prot)
{
	unsigned long offset = pfn << PAGE_SHIFT;

	if (uncached_access(file, offset))
		return pgprot_noncached(vma_prot);
	return vma_prot;
}

static int vpu_mmap(struct file *file, struct vm_area_struct *vma)
{
	size_t size = vma->vm_end - vma->vm_start;

	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	vma->vm_ops = &vpu_mmap_ops;

	if (remap_pfn_range(vma,
			    vma->vm_start,
			    vma->vm_pgoff,
			    size,
			    vma->vm_page_prot)) {
		printk(KERN_ERR	 "vpu: remap_pfn_range failed\n");
		return -EAGAIN;
	}
	return 0;
}

static struct file_operations vpu_fops = {
	.open		= vpu_open,
	.release	= vpu_release,
	.mmap		= vpu_mmap,
	.ioctl		= vpu_ioctl,
};

static void __exit vpu_cleanup(void)
{
	clk_put(vcore_clk);
	clk_put(vcache_clk);
	clk_put(vcodec_clk);
	clk_put(overlay_clk);

	unregister_chrdev(DEV_MAJOR, DEV_NAME);
}

static struct class *vpu_class;

static int __init vpu_init(void)
{
	total_cnt = 0;

	cache_droped = 0;
	memset(&opened_info, 0x00, sizeof(opened_info));
	
	vcore_clk = clk_get(NULL, "vcore");
	BUG_ON(vcore_clk == NULL);

	vcache_clk = clk_get(NULL, "vcache");
	BUG_ON(vcache_clk == NULL);
	
	vcodec_clk = clk_get(NULL, "vcodec");
	BUG_ON(vcodec_clk == NULL);

	overlay_clk = clk_get(NULL, "overlay");
	BUG_ON(overlay_clk == NULL);

	if (register_chrdev(DEV_MAJOR, DEV_NAME, &vpu_fops))
		printk(KERN_ERR "unable to get major %d for VPU device\n", DEV_MAJOR);
	vpu_class = class_create(THIS_MODULE, DEV_NAME);
	device_create(vpu_class, NULL, MKDEV(DEV_MAJOR, DEV_MINOR), NULL, DEV_NAME);
	return 0;
}

module_init(vpu_init)
module_exit(vpu_cleanup)
