#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/dma-mapping.h>
#include <linux/poll.h>
#include <linux/spi/spi.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/dma.h>

#include <bsp.h>
#include <linux/spi/tcc_tsif.h>
#include <linux/tcc_tsif_parallel_hwset.h>

#include <i-tv/itv_common.h>
#include "itcc89xx_tsif_p.h"

#define P_TSIF_SPI_NUM 1

typedef enum {
	P_TSIF_STATE_DEINIT, 
	P_TSIF_STATE_INIT, 
	P_TSIF_STATE_START, 
	P_TSIF_STATE_STOP 
} e_p_tsif_state;

typedef struct {
	e_p_tsif_state state;

	int mode;	// true : TSIF_Serial mode, false : TSIF_Parallel mode.
	struct tcc_tsif_parallel_handle handle;

	void *cb_data;
	void (*callback)(void *);
} st_p_tsif;

st_p_tsif p_tsif_inst = {
	.state = P_TSIF_STATE_DEINIT, 
};

static void tsif_p_free_dma_buffer(struct tea_dma_buf *tdma)
{
	if(tdma) {
		if(tdma->v_addr)
			dma_free_writecombine(0, tdma->buf_size, tdma->v_addr, tdma->dma_addr);
		memset(tdma, 0, sizeof(struct tea_dma_buf));
	}
}

static int tsif_p_alloc_dma_buffer(struct tea_dma_buf *tdma, unsigned int size)
{
	int error = -1;

	if(tdma) {
		tsif_p_free_dma_buffer(tdma);
		tdma->buf_size = size;
		tdma->v_addr = dma_alloc_writecombine(0, tdma->buf_size, &tdma->dma_addr, GFP_KERNEL);
		printk("[%s] Alloc DMA buffer @0x%X(Phy=0x%X), size:%d\n", __func__, (unsigned int)tdma->v_addr,
				(unsigned int)tdma->dma_addr, tdma->buf_size);

		error = tdma->v_addr ? 0 : 1;
	}

	return error;
}

static irqreturn_t tsif_p_dma_handler(int irq, void *dev_id)
{
	st_p_tsif *instance = &p_tsif_inst;
	struct tcc_tsif_parallel_handle *handle = (struct tcc_tsif_parallel_handle *)dev_id;

	handle->tsif_isr(handle);

	handle->cur_q_pos = (int)handle->dma_phy_curpos + (int)(handle->rx_dma.v_addr - handle->rx_dma.dma_addr);

	if(instance->callback)
		instance->callback(instance->cb_data);

	return IRQ_HANDLED;
	
}

int tcc_tsif_p_get_pos(void)
{
	st_p_tsif *instance = &p_tsif_inst;
	struct tcc_tsif_parallel_handle *handle = &instance->handle;

	return handle->cur_q_pos;
}
EXPORT_SYMBOL_GPL(tcc_tsif_p_get_pos);

int tcc_tsif_p_start(void)
{
	st_p_tsif *instance = &p_tsif_inst;
	struct tcc_tsif_parallel_handle *handle = &instance->handle;

	if(instance->state == P_TSIF_STATE_DEINIT)
	{
		printk("[%s] Could not start : 0x%x\n", __func__, instance->state);
		return -EACCES;
	}

	handle->msb_first		= 0x01;
	handle->big_endian		= 0x01;
	handle->serial_mode		= instance->mode;
	handle->clk_polarity	= 0x00;
	handle->valid_polarity	= 0x01;
	handle->sync_polarity	= 0x00;
	handle->sync_delay		= 0x00;
	
	handle->dma_stop(handle);
	handle->tsif_set(handle);
	handle->clear_fifo_packet(handle);
	handle->dma_start(handle);

	instance->state = P_TSIF_STATE_START;

	return 0;
}
EXPORT_SYMBOL_GPL(tcc_tsif_p_start);

int tcc_tsif_p_stop(void)
{
	st_p_tsif *instance = &p_tsif_inst;
	struct tcc_tsif_parallel_handle *handle = &instance->handle;

	if(instance->state != P_TSIF_STATE_START)
		return 0;

	handle->dma_stop(handle);

	instance->state = P_TSIF_STATE_STOP;

	return 0;
}
EXPORT_SYMBOL_GPL(tcc_tsif_p_stop);

int tcc_tsif_p_init(unsigned char **buffer_addr, unsigned int *real_buffer_size, unsigned int buffer_size, 
		unsigned int pkt_intr_cnt, unsigned int timing_mode, unsigned int pid_mode, unsigned int serial_mode, 
		void (*callback)(void *), void *cb_data) 
{
	int error = 0;

	st_p_tsif *instance = &p_tsif_inst;
	struct tcc_tsif_parallel_handle *handle = &instance->handle;
	struct tcc_tsif_regs *reg_addr;

	if(instance->state != P_TSIF_STATE_DEINIT)
	{
		printk("[%s] Already Init : 0x%x\n", __func__, instance->state);
		return -EACCES;
	}

	reg_addr = (struct tcc_tsif_regs *)((TSIF_CH == 0) ? &(HwTSIF0_BASE) : &(HwTSIF1_BASE));

	memset(instance, 0x00, sizeof(st_p_tsif));

	instance->callback = callback;
	instance->cb_data = cb_data;
	instance->mode = serial_mode;

	// reset
	tca_ckc_set_iobus_swreset(RB_TSIFCONTROLLER, OFF);
	tca_ckc_set_iobus_swreset(RB_TSIFCONTROLLER, ON);

	tca_ckc_setiobus(RB_TSIFCONTROLLER, ENABLE);

	handle->dma_intr_packet_cnt = pkt_intr_cnt;

	if(tcc_tsif_parallel_init(handle,
				reg_addr,
				tsif_p_alloc_dma_buffer,
				tsif_p_free_dma_buffer,
				TSIF_DMA_SIZE,
				TSIF_CH,
				TSIF_DMA_CONTROLLER,
				TSIF_DMA_CH
				)) {
		printk("[%s] SPI initialize failed\n", __func__);
		error = -EBUSY;
		goto err_tsif_p;
	}

	error = request_irq(handle->irq, tsif_p_dma_handler, IRQF_SHARED, 
			TSIF_DEV_NAME, handle);
	if(error) 
		goto err_irq; 

	*buffer_addr = handle->rx_dma.v_addr;
	*real_buffer_size = TSIF_DMA_SIZE;
	handle->prev_q_pos = handle->rx_dma.v_addr;

	instance->state = P_TSIF_STATE_INIT;

	return 0;

err_irq:
	free_irq(handle->irq, handle);

err_tsif_p:
	tcc_tsif_parallel_clean(handle);
	tca_ckc_setiobus(RB_TSIFCONTROLLER, DISABLE);

	return error;
}
EXPORT_SYMBOL_GPL(tcc_tsif_p_init);

void tcc_tsif_p_deinit(void)
{
	st_p_tsif *instance = &p_tsif_inst;
	struct tcc_tsif_parallel_handle *handle = &instance->handle;

	if(instance->state == P_TSIF_STATE_DEINIT)
		return;

	free_irq(handle->irq, handle);
	tcc_tsif_parallel_clean(handle);
	tca_ckc_setiobus(RB_TSIFCONTROLLER, DISABLE);

	instance->state = P_TSIF_STATE_DEINIT;
}
EXPORT_SYMBOL_GPL(tcc_tsif_p_deinit);
