/*
 * drivers/mfd/lp8725.c
 *
 * LP8725 PMIC driver
 *
 * Copyright (C) 2009 Telechips, Inc.
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 */

#include <linux/kernel.h>
#include <linux/i2c.h>
#include <linux/mfd/core.h>
#include <linux/mfd/lp8725.h>
#include <linux/regulator/lp8725.h>
#include <linux/clk.h>

#define LP8725_DRIVER_NAME "lp8725-regulator"

#define LP8725_I2C_ADDR		0x79

static unsigned short dummy[] = { I2C_CLIENT_END };

static unsigned short probe[] = { 0, LP8725_I2C_ADDR, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	.normal_i2c = dummy,
	.probe	    = probe,
	.ignore	    = dummy,
};

static struct clk *clk;

static int lp8725_attach_adapter(struct i2c_adapter *adapter);
static int lp8725_detach_client(struct i2c_client *client);

static int lp8725_read(struct lp8725 *lp8725, u8 reg, int count, u8 *data)
{
	struct i2c_client *i2c = lp8725->client;
	struct i2c_msg xfer[2];
	int ret;

	/* write register */
	xfer[0].addr = i2c->addr;
	xfer[0].flags = 0;
	xfer[0].len =  1;
	xfer[0].buf = &reg;

	/* read data */
	xfer[1].addr = i2c->addr;
	xfer[1].flags = I2C_M_RD;
	xfer[1].len = count * sizeof(u8);
	xfer[1].buf = data;

	ret = i2c_transfer(i2c->adapter, xfer, 2);
	pr_debug("%s: returns %d\n", __func__, ret);
	if (ret < 0)
		return -EINVAL;
	return 0;
}

int lp8725_write(struct lp8725 *lp8725, u8 reg, int count, u8 *data)
{
	struct i2c_client *i2c = lp8725->client;
	u8 *msg;
	int ret;

	msg = kmalloc(count * sizeof(u8) + 1, GFP_KERNEL);
	if (msg == NULL)
		return -ENOMEM;

	msg[0] = reg;
	memcpy(&msg[1], data, count * sizeof(u8));

	ret = i2c_master_send(i2c, msg, count * sizeof(u8) + 1);
	pr_debug("%s: returns %d\n", __func__, ret);
	kfree(msg);

	if (ret < 0)
		return -EINVAL;
	return 0;
}

int lp8725_reg_read(struct lp8725 *lp8725, u8 reg)
{
	u8 val;

	mutex_lock(&lp8725->lock);
	if (lp8725_read(lp8725, reg, 1, &val) < 0)
		val = -EINVAL;
	mutex_unlock(&lp8725->lock);

	return val;
}
EXPORT_SYMBOL(lp8725_reg_read);

int lp8725_reg_write(struct lp8725 *lp8725, u8 reg, u8 val)
{
	int ret;

	mutex_lock(&lp8725->lock);
	ret = lp8725_write(lp8725, reg, 1, &val);
	mutex_unlock(&lp8725->lock);

	return ret;
}
EXPORT_SYMBOL(lp8725_reg_write);

int lp8725_reg_set_bit_mask(struct lp8725 *lp8725, u8 reg, u8 mask, u8 val)
{
	int ret;
	u8 tmp;

	mutex_lock(&lp8725->lock);
	ret = lp8725_read(lp8725, reg, 1, &tmp);
	if (ret < 0)
		goto out;
	tmp &= ~mask;
	tmp |= mask & val;
	ret = lp8725_write(lp8725, reg, 1, &tmp);
out:
	mutex_unlock(&lp8725->lock);

	return ret;	
}
EXPORT_SYMBOL(lp8725_reg_set_bit_mask);

static int lp8725_init(struct lp8725 *lp8725)
{
	int i;

	/* set DCDC current limit */
	lp8725_reg_set_bit_mask(lp8725, 0x1, 0x1F, 0x19);	/* 2.80V */
	lp8725_reg_set_bit_mask(lp8725, 0x2, 0x1F, 0x19);	/* 2.80V */
	lp8725_reg_set_bit_mask(lp8725, 0x3, 0x1F, 0x1F);	/* 3.30V */
	lp8725_reg_set_bit_mask(lp8725, 0x4, 0x1F, 0x1F);	/* 3.30V */
	lp8725_reg_set_bit_mask(lp8725, 0x5, 0x1F, 0x18);	/* 2.60V */
	
	/* set LILO voltage */
	lp8725_reg_set_bit_mask(lp8725, 0x6, 0x1F, 0x8);	/* 1.20V */
	lp8725_reg_set_bit_mask(lp8725, 0x7, 0x1F, 0x10);	/* 1.80V */

	lp8725_reg_write(lp8725, 0xD, 0x7F & ~(LP8725_LDO2_EN | LP8725_LDO5_EN | LP8725_LILO2_EN));

	for (i = 0; i <= 0xf; i++) {
		pr_debug("%s: reg=0x%x, val=0x%x\n", __func__, i, lp8725_reg_read(lp8725, i));
	}

	return 0;
}

static struct i2c_driver lp8725_i2c_driver = {
	.driver		= {
		.name	= LP8725_DRIVER_NAME,
	},
//	.id		= I2C_DRIVERID_LP8725,
	.attach_adapter	= lp8725_attach_adapter,
	.detach_client	= lp8725_detach_client,
};

static int lp8725_probe(struct i2c_adapter *adapter, int addr, int kind)
{
	struct i2c_client *client;
	struct lp8725 *lp8725;
	int ret;

	client = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (client == NULL) {
		ret = -ENOMEM;
		goto err_mem;
	}

	client->addr = addr;
	client->adapter = adapter;
	client->driver = &lp8725_i2c_driver;
	strlcpy(client->name, LP8725_DRIVER_NAME, I2C_NAME_SIZE);

	ret = i2c_attach_client(client);
	if (ret < 0) {
		pr_err("lp8725: can't attach i2c client: %d\n", ret);
		goto err;
	}

	lp8725 = kzalloc(sizeof(struct lp8725), GFP_KERNEL);
	if (lp8725 == NULL) {
		ret = -ENOMEM;
		goto err;
	}

	mutex_init(&lp8725->lock);

	i2c_set_clientdata(client, lp8725);

	lp8725->client = client;

	lp8725_register_regulator(&client->dev, LP8725_DCDC1);
	lp8725_register_regulator(&client->dev, LP8725_DCDC2);
	lp8725_register_regulator(&client->dev, LP8725_LDO1);
	lp8725_register_regulator(&client->dev, LP8725_LDO2);
	lp8725_register_regulator(&client->dev, LP8725_LDO3);
	lp8725_register_regulator(&client->dev, LP8725_LDO4);
	lp8725_register_regulator(&client->dev, LP8725_LILO1);
	lp8725_register_regulator(&client->dev, LP8725_LILO2);

	ret = lp8725_init(lp8725);
	if (ret != 0)
		goto err;

	return 0;
err:
	kfree(client);
err_mem:
	return ret;
}

static int lp8725_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter, &addr_data, lp8725_probe);
}

static int lp8725_detach_client(struct i2c_client *client)
{
	return i2c_detach_client(client);
}

static int __init lp8725_module_init(void)
{
	int ret;

	pr_info("LP8725 regulator driver init\n");

	clk = clk_get(NULL, "i2c");
	if (clk) {
		clk_enable(clk);
	} else {
		pr_err("lp8725: can't get i2c clock\n");
		return -ENOMEM;
	}

	ret = i2c_add_driver(&lp8725_i2c_driver);
	if (ret != 0) {
		pr_err("lp8725: failed to register driver: %d\n", ret);
	}
	return ret;
}
module_init(lp8725_module_init);

static void __exit lp8725_module_exit(void)
{
	i2c_del_driver(&lp8725_i2c_driver);
	clk_disable(clk);
	clk_put(clk);
}
module_exit(lp8725_module_exit);

MODULE_AUTHOR("Telechips, Inc.");
MODULE_LICENSE("GPL");
