/*
 * linux/drivers/spi/tsif_tcc79x.c
 *
 * Author:  <linux@telechips.com>
 * Created: 10th Nov, 2008 
 * Description: Driver for TCC79X SPI-Slave Controllers
 *
 * Copyright (c) Telechips, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/platform_device.h>

#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/dma-mapping.h>
#include <asm/uaccess.h>

#include <linux/spi/spi.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/hardware.h>

#include <linux/workqueue.h>
#include <linux/poll.h>

#include <asm/arch/tcc79x.h>
#include <asm/plat-tcc/tcc_ckc.h>

#include <linux/spi/tsif_tcc79x.h>
#include "tca_spi_hwset.h"

struct tca_spi_pri_handle {
    wait_queue_head_t wait_q;
    struct mutex mutex;
    int open_cnt;
};

static tca_spi_handle_t tsif_handle;
static struct tca_spi_pri_handle tsif_pri;

static int __init tsif_drv_probe(struct platform_device *dev)
{
    platform_set_drvdata(dev, &tsif_handle);
    return 0;
}

static int tsif_drv_suspend(struct platform_device *dev, pm_message_t state)
{
	return 0;
}

static int tsif_drv_resume(struct platform_device *dev)
{
    tca_spi_handle_t *p_tsif = platform_get_drvdata(dev);

    if (p_tsif) {
        if (p_tsif->id == 0) {
            HwGPSB_PCFG = (HwGPSB_PCFG & ~0x000000FF) | (0x6);
        } else {
            HwGPSB_PCFG = (HwGPSB_PCFG & ~0x0000FF00) | (0x500);
        }
        if (p_tsif->is_enable_dma(p_tsif)) {
            p_tsif->dma_stop(p_tsif);
            p_tsif->dma_start(p_tsif);
        }
    }

	return 0;
}

static struct platform_driver tsif_platform_driver = {
	.driver	= {
		.name    = "tsif",
		.owner	 = THIS_MODULE,
	},
	.suspend = tsif_drv_suspend,
	.resume  = tsif_drv_resume,
};


static void tea_free_dma_linux(struct tea_dma_buf *tdma)
{
    if (tdma) {
        if (tdma->dma_addr != 0) {
            dma_free_writecombine(0, tdma->buf_size, tdma->v_addr, tdma->dma_addr);
        }
        memset(tdma, 0, sizeof(struct tea_dma_buf));
    }
}

static int tea_alloc_dma_linux(struct tea_dma_buf *tdma, unsigned int size)
{
    int ret = -1;
    if (tdma) {
        tea_free_dma_linux(tdma);
        tdma->buf_size = size;
        tdma->v_addr = dma_alloc_writecombine(0, tdma->buf_size, &tdma->dma_addr, GFP_KERNEL);
        printk("Malloc DMA buffer @0x%X(Phy=0x%X), size:%d\n",
               (unsigned int)tdma->v_addr,
               (unsigned int)tdma->dma_addr,
               tdma->buf_size);
        ret = tdma->v_addr ? 0 : 1;
    }
    return ret;
}


static irqreturn_t tcc79x_tsif_dma_handler(int irq, void *dev_id)
{
    struct tca_spi_handle *tspi = (struct tca_spi_handle *)dev_id;
    struct tca_spi_pri_handle *tpri = (struct tca_spi_pri_handle *)tspi->private_data;
    unsigned long dma_done_reg = 0;

    dma_done_reg = tspi->regs->DMAICR;
    if (dma_done_reg & Hw28) {
        BITSET(tspi->regs->DMAICR, Hw29 | Hw28);
        if (tpri->open_cnt > 0) {
            tspi->cur_q_pos = (int)(tspi->regs->DMASTAT >> 17);
            wake_up(&(tpri->wait_q));
        }
    }
    return IRQ_HANDLED;
}

static int tsif_get_readable_cnt(tca_spi_handle_t *H)
{
    if (H) {
        //int dma_pos = (int)(H->regs->DMASTAT >> 17);
        int dma_pos = H->cur_q_pos;
        int readable_cnt = 0;

        if (dma_pos > H->q_pos) {
            readable_cnt = dma_pos - H->q_pos;
        } else if (dma_pos < H->q_pos) {
            readable_cnt = H->dma_total_packet_cnt - H->q_pos;
            readable_cnt += dma_pos;
        } 
        return readable_cnt;
    }
    return 0;
}


static ssize_t tcc79x_tsif_read(struct file *filp, char *buf, size_t len, loff_t *ppos)
{
    int readable_cnt = 0, copy_cnt = 0;
    int copy_byte = 0;

    readable_cnt = tsif_get_readable_cnt(&tsif_handle);
    if (readable_cnt > 0) {
        copy_byte = readable_cnt * TSIF_PACKET_SIZE;
        if (copy_byte > len) {
            copy_byte = len;
        }

        copy_byte -= copy_byte % TSIF_PACKET_SIZE;
        copy_cnt = copy_byte / TSIF_PACKET_SIZE;
        if (copy_cnt >= tsif_handle.dma_intr_packet_cnt) {
            int offset = tsif_handle.q_pos * TSIF_PACKET_SIZE;
            if (copy_cnt > tsif_handle.dma_total_packet_cnt - tsif_handle.q_pos) {
                int first_copy_byte = tsif_handle.dma_total_packet_cnt - tsif_handle.q_pos;
                int first_copy_cnt = first_copy_byte / TSIF_PACKET_SIZE;
                int second_copy_byte = (copy_cnt - first_copy_cnt) * TSIF_PACKET_SIZE;

                if (copy_to_user(buf, tsif_handle.rx_dma.v_addr + offset, first_copy_byte)) {
                    return -EFAULT;
                }
                if (copy_to_user(buf + first_copy_byte, tsif_handle.rx_dma.v_addr, second_copy_byte)) {
                    return -EFAULT;
                }

                tsif_handle.q_pos = copy_cnt - first_copy_cnt;
            } else {
                if (copy_to_user(buf, tsif_handle.rx_dma.v_addr + offset, copy_byte)) {
                    return -EFAULT;
                }

                tsif_handle.q_pos += copy_cnt;
                if (tsif_handle.q_pos >= tsif_handle.dma_total_packet_cnt) {
                    tsif_handle.q_pos = 0;
                }
            }
            return copy_byte;
        }
    }
    return 0;
}

static ssize_t tcc79x_tsif_write(struct file *filp, const char *buf, size_t len, loff_t *ppos)
{
    return 0;
}

static unsigned int tcc79x_tsif_poll(struct file *filp, struct poll_table_struct *wait)
{
    if (tsif_get_readable_cnt(&tsif_handle) >= tsif_handle.dma_intr_packet_cnt) {
		return  (POLLIN | POLLRDNORM);
    }

    poll_wait(filp, &(tsif_pri.wait_q), wait);
    if (tsif_get_readable_cnt(&tsif_handle) >= tsif_handle.dma_intr_packet_cnt) {
		return  (POLLIN | POLLRDNORM);
    }
    return 0;
}

static int tcc79x_tsif_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    int ret = 0;
    switch (cmd) {
    case IOCTL_TSIF_DMA_START:
        {
            struct tcc_tsif_param param;
            if (copy_from_user(&param, (void *)arg, sizeof(struct tcc_tsif_param))) {
                return -EFAULT;
            }

            if (((TSIF_PACKET_SIZE * param.ts_total_packet_cnt) > TSIF_DTCM_DMA_MEM_SIZE)
                || (param.ts_total_packet_cnt <= 0)) {
                return -EFAULT;
            }

            tsif_handle.dma_stop(&tsif_handle);

            tca_spi_setCPHA(tsif_handle.regs, param.mode & SPI_CPHA);
            tca_spi_setCPOL(tsif_handle.regs, param.mode & SPI_CPOL);
            tca_spi_setCS_HIGH(tsif_handle.regs, param.mode & SPI_CS_HIGH);
            tca_spi_setLSB_FIRST(tsif_handle.regs, param.mode & SPI_LSB_FIRST);

            tsif_handle.dma_total_packet_cnt = param.ts_total_packet_cnt;
            tsif_handle.dma_intr_packet_cnt = param.ts_intr_packet_cnt;

            tsif_handle.clear_fifo_packet(&tsif_handle);
            tsif_handle.q_pos = tsif_handle.cur_q_pos = 0;

            tsif_handle.set_packet_cnt(&tsif_handle, MPEG_PACKET_SIZE);
            tsif_handle.dma_start(&tsif_handle);
        }
        break;
    case IOCTL_TSIF_DMA_STOP:
            tsif_handle.dma_stop(&tsif_handle);
        break;
    case IOCTL_TSIF_GET_MAX_DMA_SIZE:
        {
            struct tcc_tsif_param param;

            param.ts_total_packet_cnt = TSIF_DTCM_DMA_MEM_SIZE / TSIF_PACKET_SIZE;
            param.ts_intr_packet_cnt = 1;

            if (copy_to_user((void *)arg, (void *)&param, sizeof(struct tcc_tsif_param))) {
                return -EFAULT;
            }
        }
        break;
    case IOCTL_TSIF_SET_PID:
        {
            struct tcc_tsif_pid_param param;
            if (copy_from_user(&param, (void *)arg, sizeof(struct tcc_tsif_pid_param))) {
                return -EFAULT;
            }

            if (param.valid_data_cnt <= PID_MATCH_TABLE_MAX_CNT) {
                int i = 0;
                for (i = 0; i < PID_MATCH_TABLE_MAX_CNT; i++) {
                    HwGPSB_PIDT(i) = 0;
                }
                if (param.valid_data_cnt > 0) {
                    for (i = 0; i < param.valid_data_cnt; i++) {
                        HwGPSB_PIDT(i) = (param.pio_data)[i] & 0x3FFFFFFF;
                        BITSET(HwGPSB_PIDT(i), (TSIF_SPI_NUM == 0) ? HwGPSB_PIDT_CH0 : HwGPSB_PIDT_CH1);
                    }
                    tsif_handle.set_mpegts_pidmode(&tsif_handle, 1);
                } else {
                    tsif_handle.set_mpegts_pidmode(&tsif_handle, 0);
                }
            } else {
                printk("tsif: PID TABLE is so big !!! (0x%X)\n", cmd);
                ret = -EINVAL;
            }
        }
        break;
    default:
        printk("tsif: unrecognized ioctl (0x%X)\n", cmd);
        ret = -EINVAL;
        break;
    }
    return ret;
}

static int tcc79x_tsif_open(struct inode *inode, struct file *filp)
{
    int ret = 0;
    mutex_lock(&(tsif_pri.mutex));
    if (tsif_pri.open_cnt == 0) {
        tsif_pri.open_cnt++;
    } else {
        ret = -EBUSY;
    }
    mutex_unlock(&(tsif_pri.mutex));

    tsif_handle.set_mpegts_pidmode(&tsif_handle, 0);
    return ret;
}

static int tcc79x_tsif_release(struct inode *inode, struct file *filp)
{
    mutex_lock(&(tsif_pri.mutex));
    tsif_handle.dma_stop(&tsif_handle);
    if (tsif_pri.open_cnt > 0) {
        tsif_pri.open_cnt--;
    }
    mutex_unlock(&(tsif_pri.mutex));
    return 0;
}

struct file_operations tcc79x_tsif_fops =
{
    .owner      = THIS_MODULE,
    .read       = tcc79x_tsif_read,
    .write      = tcc79x_tsif_write,
    .ioctl      = tcc79x_tsif_ioctl,
    .open       = tcc79x_tsif_open,
    .release    = tcc79x_tsif_release,
    .poll       = tcc79x_tsif_poll,
};

static int __init tsif_init(void)
{
    int ret;
    volatile struct tca_spi_regs *reg_addr = NULL;

    printk("[%s:%d] TSIF start \n", __func__, __LINE__); 
    reg_addr = (volatile struct tca_spi_regs *)((TSIF_SPI_NUM == 0)
                                                ? &(HwGPSB_CH0_BASE) : &(HwGPSB_CH1_BASE));

    memset(&tsif_handle, 0, sizeof(tca_spi_handle_t));
    memset(&tsif_pri, 0, sizeof(struct tca_spi_pri_handle));
    ret = register_chrdev(TSIF_DEV_MAJOR, TSIF_DEV_NAME, &tcc79x_tsif_fops);
    if (ret < 0) {
        printk("[%s:%d] register_chrdev error !!!!!\n", __func__, __LINE__); 
        return ret;
    }

    tca_ckc_com_setperibus(TSIF_SPI_NUM == 0 ? RB_GPSB0 : RB_GPSB1, 1);
    tca_ckc_com_setperi(TSIF_SPI_NUM == 0 ? PERI_GPSB0 : PERI_GPSB1, KHz, 40000, DIRECTPLL0);

    if (tca_spi_init(&tsif_handle,
                     reg_addr,
                     INT_GPSB1_DMA,
                     tea_alloc_dma_linux,
                     NULL,
                     TSIF_SPI_NUM,
                     1)) {
        printk("[%s:%d] tca_spi_init error !!!!!\n", __func__, __LINE__); 
        return -EINVAL;
    }

    tsif_handle.private_data = (void *)&tsif_pri;
    tsif_handle.clear_fifo_packet(&tsif_handle);
    tsif_handle.dma_stop(&tsif_handle);

    init_waitqueue_head(&(tsif_pri.wait_q));
    mutex_init(&(tsif_pri.mutex));

    tsif_handle.dma_total_packet_cnt = TSIF_DTCM_DMA_MEM_SIZE / TSIF_PACKET_SIZE;
    tsif_handle.dma_intr_packet_cnt = 1;

    tsif_handle.hw_init(&tsif_handle);
    ret = request_irq(tsif_handle.irq, tcc79x_tsif_dma_handler, IRQF_SHARED, "tsif", &tsif_handle);

    tsif_handle.set_packet_cnt(&tsif_handle, MPEG_PACKET_SIZE);

    ret =  platform_driver_probe(&tsif_platform_driver, tsif_drv_probe);

    return 0;
}

static void __exit tsif_exit(void)
{
    free_irq(tsif_handle.irq, &tsif_handle);

    tca_spi_clean(&tsif_handle);

    tca_ckc_com_setperibus(TSIF_SPI_NUM == 0 ? RB_GPSB0 : RB_GPSB1, 0);

    unregister_chrdev(TSIF_DEV_MAJOR, TSIF_DEV_NAME);

    platform_driver_unregister(&tsif_platform_driver);
}

module_init(tsif_init);
module_exit(tsif_exit);

MODULE_AUTHOR("Telechips Inc. SYS4-3 linux@telechips.com");
MODULE_DESCRIPTION("TCC79x TSIF driver");
MODULE_LICENSE("GPL");

