/*
 * linux/drivers/usb/host/ohci-tcc.c
 *
 * Description: OHCI HCD (Host Controller Driver) for USB.
 *              Bus Glue for Telechips TCCXXX
 *
 * (C) Copyright 1999 Roman Weissgaerber <weissg@vienna.at>
 * (C) Copyright 2000-2002 David Brownell <dbrownell@users.sourceforge.net>
 * (C) Copyright 2002 Hewlett-Packard Company
 * (C) Copyright 2008-2009 Telechips <linux@telechips.com>
 * (C) Copyright 2008-2009 Telechips 
 *
 * Rewritten by <linux@telechips.com>
 * Written by Christopher Hoover <ch@hpl.hp.com>
 * Based on fragments of previous driver by Russell King et al.
 *
 * Modified for LH7A404 from ohci-sa1111.c
 *  by Durgesh Pattamatta <pattamattad@sharpsec.com>
 *
 * Modified for pxa27x from ohci-lh7a404.c
 *  by Nick Bane <nick@cecomputing.co.uk> 26-8-2004
 *
 * Modified for TCCXXX from ohci-pxa27x.c
 *  by Telechips Linux Team <linux@telechips.com> 03-9-2008
 *
 * This file is licenced under the GPL.
 */

#include <linux/device.h>
#include <linux/signal.h>
#include <linux/platform_device.h>
#include <asm/mach-types.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <mach/bsp.h>
#include <mach/ohci.h>
#include <mach/tcc_pca953x.h>

static int tcc_ohci_select_pmm(unsigned long reg_base, int mode, int num_port, int *port_mode)
{
	volatile struct _USBHOST11 *ohci_reg;
	ohci_reg = (volatile struct _USBHOST11 *)reg_base;

	switch (mode) {
	case USBOHCI_PPM_NPS:
		// set NO Power Switching mode
        ohci_reg->HcRhDescriptorA |= USBOHCI_UHCRHDA_NPS;
		break;
		
	case USBOHCI_PPM_GLOBAL:
		// make sure the NO Power Switching mode bit is OFF so Power Switching can occur
        // make sure the PSM bit is CLEAR, which allows all ports to be controlled with
        // the GLOBAL set and clear power commands
        ohci_reg->HcRhDescriptorA &= ~(USBOHCI_UHCRHDA_NPS|USBOHCI_UHCRHDA_PSM_PERPORT);
		break;
		
	case USBOHCI_PPM_PERPORT:
		// make sure the NO Power Switching mode bit is OFF so Power Switching can occur
        // make sure the PSM bit is SET, which allows all ports to be controlled with
        // the PER PORT set and clear power commands
        ohci_reg->HcRhDescriptorA &= ~USBOHCI_UHCRHDA_NPS;
        ohci_reg->HcRhDescriptorA |=  USBOHCI_UHCRHDA_PSM_PERPORT;

		// set the power management mode for each individual port to Per Port.
        {
            int p;
            for (p = 0; p < num_port; p++) {
                ohci_reg->HcRhDescriptorB |= (unsigned int)( 1u << (p+17) );   // port 1 begins at bit 17
            }
        }
		break;
		
    case USBOHCI_PPM_MIXED:
        // make sure the NO Power Switching mode bit is OFF so Power Switching can occur
        // make sure the PSM bit is SET, which allows all ports to be controlled with
        // the PER PORT set and clear power commands
        ohci_reg->HcRhDescriptorA &= ~USBOHCI_UHCRHDA_NPS;
        ohci_reg->HcRhDescriptorA |=  USBOHCI_UHCRHDA_PSM_PERPORT;
		
        // set the power management mode for each individual port to Per Port.
        // if the value in the PortMode array is non-zero, set Per Port mode for the port.
        // if the value in the PortMode array is zero, set Global mode for the port
        {
			int p;
			for (p = 0; p < num_port; p++) {
				if (port_mode[p]) {
					ohci_reg->HcRhDescriptorB |= (unsigned int)( 1u << (p+17) );   // port 1 begins at bit 17
				} else {
					ohci_reg->HcRhDescriptorB &= ~(unsigned int)( 1u << (p+17) );  // port 1 begins at bit 17
				}
			}
        }
		
        break;
		
	default:
		printk(KERN_ERR "Invalid mode %d, set to non-power switch mode.\n", mode);
		ohci_reg->HcRhDescriptorA |= USBOHCI_UHCRHDA_NPS;
	}

	return 0;
}

extern int usb_disabled(void);

/*-------------------------------------------------------------------------*/

static void USBHostVBUSControl(int on)
{	
	/* GPIO_EXPAND HVBUS & PWR_GP1 control */
	if (on) {
		tcc_pca953x_setup(PCA9538_U4_SLAVE_ADDR, PWR_GP1, OUTPUT, HIGH, SET_DIRECTION|SET_VALUE);
		tcc_pca953x_setup(PCA9538_U4_SLAVE_ADDR, HVBUS_ON, OUTPUT, HIGH, SET_DIRECTION|SET_VALUE);
	} else {
#if !defined(CONFIG_TCC_DWC_OTG) && !defined(CONFIG_TCC_DWC_OTG_MODULE)
		tcc_pca953x_setup(PCA9538_U4_SLAVE_ADDR, PWR_GP1, OUTPUT, LOW, SET_DIRECTION|SET_VALUE);
#endif
		tcc_pca953x_setup(PCA9538_U4_SLAVE_ADDR, HVBUS_ON, OUTPUT, LOW, SET_DIRECTION|SET_VALUE);
	}
}

void TurnOnUSBHostClocks(void)
{
	PIOBUSCFG pIOBUSCfg;
	pIOBUSCfg = (PIOBUSCFG)tcc_p2v(HwIOBUSCFG_BASE);
// JINS 2009.11.18	tca_ckc_setperi(PERI_USB11H, ENABLE, 480000, PCDIRECTPLL2);
	tca_ckc_setperi(PERI_USB11H, ENABLE, 480000, PCDIRECTPLL1);
	BITSET(pIOBUSCfg->HCLKEN0, Hw0);	// Bit0 is USB1.1 Host IOBUS AHB clock
	BITSET(pIOBUSCfg->HRSTEN0, Hw0);	// Bit0 is USB1.1 Host IOBUS AHB Hreset
}

void TurnOffUSBHostClocks(void)
{
    PIOBUSCFG pIOBUSCfg;
	pIOBUSCfg = (PIOBUSCFG)tcc_p2v(HwIOBUSCFG_BASE);
// JINS 2009.11.18	tca_ckc_setperi(PERI_USB11H, DISABLE, 480000, PCDIRECTPLL2);
	tca_ckc_setperi(PERI_USB11H, DISABLE, 480000, PCDIRECTPLL1);
	BITCLR(pIOBUSCfg->HCLKEN0, Hw0);
	BITCLR(pIOBUSCfg->HRSTEN0, Hw0);
}

void USBHostReset(void)
{
	PIOBUSCFG pIOBUSCfg;
	pIOBUSCfg = (PIOBUSCFG)tcc_p2v(HwIOBUSCFG_BASE);
	BITCLR(pIOBUSCfg->HRSTEN0, Hw0);
	BITSET(pIOBUSCfg->HRSTEN0, Hw0);
	
	// USBH config
	//HwUPCFG0 = 0x800008b0; //800008b0
	//HwUPCFG2 = 0x00005500; //12MHz
	//HwUPCFG1 = 0x3634;
	//HwUPCFG3 = 0x9040;
}

int tcc_set_cken(int sw)
{
	if (sw){
	    TurnOnUSBHostClocks();
		USBHostReset();
	} else {
		TurnOffUSBHostClocks();
	}
	return 0;
}

static int tcc_start_hc(struct device *dev)
{
	int retval = 0;
	struct tccohci_platform_data *inf;
	PIOBUSCFG pIOBUSCfg;
	inf = dev->platform_data;

	/* USB 1.1 Host Configuration Register
	 * DP, DN pull-down enable
	 */
	pIOBUSCfg = (PIOBUSCFG)tcc_p2v(HwIOBUSCFG_BASE);
	BITSET(pIOBUSCfg->USB11H, Hw4|Hw3);

	tcc_set_cken(1);

	if (inf->init)
		retval = inf->init(dev);
	if (retval < 0)
		return retval;

	return 0;
}

static void tcc_stop_hc(struct device *dev)
{
	struct tccohci_platform_data *inf;
	inf = dev->platform_data;
	if (inf->exit)
		inf->exit(dev);

	tcc_set_cken(0);
}

/*-------------------------------------------------------------------------*/

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */


/**
 * usb_hcd_tcc_probe - initialize tcc-based HCDs
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 */
int usb_hcd_tcc_probe (const struct hc_driver *driver, struct platform_device *pdev)
{
	int retval = 0;
	struct usb_hcd *hcd;
	struct tccohci_platform_data *inf;
	inf = pdev->dev.platform_data;

	if (!inf)
		return -ENODEV;

	if (pdev->resource[1].flags != IORESOURCE_IRQ) {
		pr_debug ("resource[1] is not IORESOURCE_IRQ");
		return -ENOMEM;
	}

	hcd = usb_create_hcd (driver, &pdev->dev, "tcc");
	if (!hcd)
		return -ENOMEM;
	hcd->rsrc_start = io_p2v(pdev->resource[0].start);
	hcd->rsrc_len   = pdev->resource[0].end - pdev->resource[0].start + 1;

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed");
		retval = -EBUSY;
		goto err1;
	}
	hcd->regs = (void __iomem *)(int)(hcd->rsrc_start);

	//TODO: Expand GPIO VBUS ON
#ifdef CONFIG_MACH_TCC8900
	/* GPIO_EXPAND HVBUS & PWR_GP1 Power-on */
	USBHostVBUSControl(1);
	retval = tcc_pca953x_setup(PCA9538_U4_SLAVE_ADDR, 0, OUTPUT, 0, GET_VALUE);
	if (!(retval & PWR_GP1) || !(retval & HVBUS_ON)) {
		printk(KERN_ERR "ohci-tcc: GPIO_EXPAND power failed\n");
		retval = -EIO;
		goto err2;
	}
#endif	


	if ((retval = tcc_start_hc(&pdev->dev)) < 0) {
		pr_debug("tcc_start_hc failed");
		goto err2;
	}

	/* Select Power Management Mode */
	tcc_ohci_select_pmm(hcd->rsrc_start, inf->port_mode, 0, 0);

	if (inf->power_budget)
		hcd->power_budget = inf->power_budget;

	ohci_hcd_init(hcd_to_ohci(hcd));

	retval = usb_add_hcd(hcd, pdev->resource[1].start, IRQF_DISABLED);
	if (retval == 0)
		return retval;

	tcc_stop_hc(&pdev->dev);
err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
err1:
	usb_put_hcd(hcd);
	return retval;
}


/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_tcc_remove - shutdown processing for tcc-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_tcc_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
void usb_hcd_tcc_remove (struct usb_hcd *hcd, struct platform_device *pdev)
{
	usb_remove_hcd(hcd);
	tcc_stop_hc(&pdev->dev);

#ifdef CONFIG_MACH_TCC8900
	/* GPIO_EXPAND HVBUS & PWR_GP1 Power-off */
	USBHostVBUSControl(0);
#endif

	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
}

/*-------------------------------------------------------------------------*/

static int __devinit
ohci_tcc_start (struct usb_hcd *hcd)
{
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int ret;

	tcc_set_cken(1);

	/* The value of NDP in roothub_a is incorrect on this hardware */
	ohci->num_ports = 1;

	if ((ret = ohci_init(ohci)) < 0)
	{
		return ret;
	}

	if ((ret = ohci_run (ohci)) < 0)
	{
		err ("can't start %s", hcd->self.bus_name);
		ohci_stop (hcd);
		return ret;
	}

	return 0;
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ohci_tcc_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"TCC OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			ohci_irq,
	.flags =		HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.start =		ohci_tcc_start,
	.stop =			ohci_stop,
#ifndef __OLD__
	.shutdown =		ohci_shutdown,
#endif

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
#ifdef  CONFIG_PM
	.bus_suspend = 		ohci_bus_suspend,
	.bus_resume =		ohci_bus_resume,
#endif
	.start_port_reset =	ohci_start_port_reset,
};

/*-------------------------------------------------------------------------*/

static int ohci_hcd_tcc_drv_probe(struct platform_device *pdev)
{
	if (usb_disabled())
		return -ENODEV;

	return usb_hcd_tcc_probe(&ohci_tcc_hc_driver, pdev);
}

static int ohci_hcd_tcc_drv_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	usb_hcd_tcc_remove(hcd, pdev);
	platform_set_drvdata(pdev, NULL);
	return 0;
}

#ifdef	CONFIG_PM
static int ohci_hcd_tcc_drv_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	if (time_before(jiffies, ohci->next_statechange))
		msleep(5);
	ohci->next_statechange = jiffies;

	tcc_stop_hc(&pdev->dev);
#ifdef __OLD__
	hcd->state = HC_STATE_SUSPENDED;	//old version
#endif
	pdev->dev.power.power_state = PMSG_SUSPEND;

	return 0;
}

static int ohci_hcd_tcc_drv_resume(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int status;
	if (time_before(jiffies, ohci->next_statechange))
		msleep(5);
	ohci->next_statechange = jiffies;

	if ((status = tcc_start_hc(&pdev->dev)) < 0)
		return status;

#ifdef __OLD__
	pdev->dev.power.power_state = PMSG_ON;	//old version
	usb_hcd_resume_root_hub(hcd);
#else
	ohci_finish_controller_resume(hcd);
#endif

	return 0;
}
#endif


static struct platform_driver ohci_hcd_tcc_driver = {
	.probe		= ohci_hcd_tcc_drv_probe,
	.remove		= ohci_hcd_tcc_drv_remove,
#ifdef CONFIG_PM
	.suspend	= ohci_hcd_tcc_drv_suspend,
	.resume		= ohci_hcd_tcc_drv_resume,
#endif
	.driver		= {
		.name	= "tcc-ohci",
		.owner	= THIS_MODULE,
	},
};

/* work with hotplug and coldplug */
MODULE_ALIAS("platform:tcc-ohci");
